
import UIKit

class Hipsterize {
  let image: UIImage

  init(image: UIImage) {
    self.image = Hipsterize.normalizedImage(image)
  }

  func hipsterElements() -> [HipsterElement] {
    var elements = [HipsterElement]()

    let ciImage = CIImage(image: image)

    let context = CIContext()
    let detector = CIDetector(ofType: CIDetectorTypeFace, context: context, options: [CIDetectorAccuracy: CIDetectorAccuracyHigh])
    let faces = detector.featuresInImage(ciImage)

    var counter = 0
    for face in faces as [CIFaceFeature] {
      let faceBounds = convertedRectForImage(image, originalRect: face.bounds)
      let faceWidth = CGRectGetWidth(face.bounds)
      var eyeAngle: CGFloat = 0

      if (face.hasLeftEyePosition && face.hasRightEyePosition) {
        let rightEyePosition = convertedPointForImage(image, originalPoint: face.rightEyePosition)
        let leftEyePosition = convertedPointForImage(image, originalPoint: face.leftEyePosition)

        let eyeCenter = CGPoint(x: (rightEyePosition.x + leftEyePosition.x) / 2.0, y: (rightEyePosition.y + leftEyePosition.y) / 2.0)

        eyeAngle = atan2(rightEyePosition.y - leftEyePosition.y, rightEyePosition.x - leftEyePosition.x)

        // Reminder: HipsterElement wants relative percentage values
        elements.append(HipsterElement(
          type: counter % 2 == 0 ? .Glasses1 : .Glasses2,
          relativeCenter: CGPoint(x: eyeCenter.x / image.size.width, y: eyeCenter.y / image.size.height),
          relativeWidth: faceWidth / image.size.width,
          rotation: eyeAngle))

        counter++
      }

      let faceTopCenter = CGPoint(x: faceBounds.origin.x + CGRectGetWidth(faceBounds)/2, y: faceBounds.origin.y - CGRectGetHeight(faceBounds)/4)
      let faceCenter = CGPoint(x: faceBounds.origin.x + CGRectGetWidth(faceBounds)/2, y: faceBounds.origin.y + CGRectGetHeight(faceBounds)/2)

      let s = sin(eyeAngle)
      let c = cos(eyeAngle)
      let rotatedCenter = CGPoint(x: (c * (faceTopCenter.x-faceCenter.x) - s * (faceTopCenter.y-faceCenter.y)) + faceCenter.x,
        y: (s * (faceTopCenter.x-faceCenter.x) + c * (faceTopCenter.y-faceCenter.y)) + faceCenter.y)

      elements.append(HipsterElement(
        type: .Fedora1,
        relativeCenter: CGPoint(x: rotatedCenter.x / image.size.width, y: rotatedCenter.y / image.size.height),
        relativeWidth: faceWidth*1.2 / image.size.width,
        rotation: 0))
    }

    return elements
  }

  func imageWithRenderedElements(elements: [HipsterElement]) -> UIImage {
    let imageView = UIImageView(image: image)

    for element in elements {
      if element.active {
        element.createViewForImageFrame(imageView.frame)
        imageView.addSubview(element.view)
      }
    }

    UIGraphicsBeginImageContextWithOptions(imageView.frame.size, false, 0.0)
    imageView.layer.renderInContext(UIGraphicsGetCurrentContext())
    let processedImage = UIGraphicsGetImageFromCurrentImageContext()
    UIGraphicsEndImageContext()
    return processedImage
  }

  private func convertedRectForImage(image: UIImage, originalRect: CGRect) -> CGRect {
    var transform = CGAffineTransformMakeScale(1, -1)
    transform = CGAffineTransformTranslate(transform, 0, -image.size.height);

    return CGRectApplyAffineTransform(originalRect, transform)
  }

  private func convertedPointForImage(image: UIImage, originalPoint: CGPoint) -> CGPoint {
    var convertedPoint = originalPoint
    var imageWidth = image.size.width
    var imageHeight = image.size.height

    switch (image.imageOrientation) {
    case .Up:
      convertedPoint.x = originalPoint.x;
      convertedPoint.y = imageHeight - originalPoint.y;
    case .Down:
      convertedPoint.x = imageWidth - originalPoint.x;
      convertedPoint.y = originalPoint.y;
    case .Left:
      convertedPoint.x = imageWidth - originalPoint.y;
      convertedPoint.y = imageHeight - originalPoint.x;
    case .Right:
      convertedPoint.x = originalPoint.y;
      convertedPoint.y = originalPoint.x;
    case .UpMirrored:
      convertedPoint.x = imageWidth - originalPoint.x;
      convertedPoint.y = imageHeight - originalPoint.y;
    case .DownMirrored:
      convertedPoint.x = originalPoint.x;
      convertedPoint.y = originalPoint.y;
    case .LeftMirrored:
      convertedPoint.x = imageWidth - originalPoint.y;
      convertedPoint.y = originalPoint.x;
    case .RightMirrored:
      convertedPoint.x = originalPoint.y;
      convertedPoint.y = imageHeight - originalPoint.x;
    }
    return convertedPoint
  }

  private class func normalizedImage(image: UIImage) -> UIImage {
    if image.imageOrientation == .Up {
      return image
    }

    UIGraphicsBeginImageContextWithOptions(image.size, false, image.scale)
    image.drawInRect(CGRect(origin: CGPoint(x: 0, y: 0), size: image.size))
    let normalizedImage = UIGraphicsGetImageFromCurrentImageContext()
    UIGraphicsEndImageContext()
    return normalizedImage
  }
}
