
import UIKit
import CoreImage
import ImageIO

class HipsterizeViewController: UIViewController {
  @IBOutlet weak var imageView: UIImageView!
  @IBOutlet weak var paletteView: UIView!

  @IBOutlet weak var glasses1: UIImageView!
  @IBOutlet weak var glasses2: UIImageView!
  @IBOutlet weak var fedora1: UIImageView!

  var image: UIImage?
  var hipsterElements = [HipsterElement]()
  var selectedElement: HipsterElement?

  var pinchRecognizer: UIPinchGestureRecognizer!
  var rotationRecognizer: UIRotationGestureRecognizer!
  var tapRecognizer: UITapGestureRecognizer!
  var panRecognizer: UIPanGestureRecognizer!

  var editsWereMade = false

  /// Get the frame of the image inside an aspect fit image view
  var imageFrame: CGRect {
    if let image = imageView.image {
      let imageScale = min(CGRectGetWidth(imageView.frame) / image.size.width, CGRectGetHeight(imageView.frame) / image.size.height)
      let scaledImageSize = CGSize(width: image.size.width * imageScale, height: image.size.height * imageScale)
      let theFrame = CGRect(x: (CGRectGetWidth(imageView.bounds) - scaledImageSize.width) / 2.0,
        y: (CGRectGetHeight(imageView.bounds) - scaledImageSize.height) / 2.0,
        width: scaledImageSize.width,
        height: scaledImageSize.height)
      return view.convertRect(theFrame, fromView: imageView)
    } else {
      return CGRect.nullRect
    }
  }

  // MARK: - UIViewController

  override func viewDidLoad() {
    super.viewDidLoad()

    panRecognizer = UIPanGestureRecognizer(target: self, action: Selector("handlePan:"))
    view.addGestureRecognizer(panRecognizer)

    tapRecognizer = UITapGestureRecognizer(target: self, action: Selector("handleTap:"))
    tapRecognizer.delegate = self
    view.addGestureRecognizer(tapRecognizer)

    // Only setting the delegate for these two so they can run simultaneously
    pinchRecognizer = UIPinchGestureRecognizer(target: self, action: Selector("handlePinch:"))
    pinchRecognizer.delegate = self
    view.addGestureRecognizer(pinchRecognizer)

    rotationRecognizer = UIRotationGestureRecognizer(target: self, action: Selector("handleRotation:"))
    rotationRecognizer.delegate = self
    view.addGestureRecognizer(rotationRecognizer)
  }

  override func viewWillTransitionToSize(size: CGSize, withTransitionCoordinator coordinator: UIViewControllerTransitionCoordinator) {
    super.viewWillTransitionToSize(size, withTransitionCoordinator: coordinator)

    deselectElement()
    removePaletteElements()

    for element in hipsterElements {
      element.view.hidden = true
    }

    coordinator.animateAlongsideTransition({ (context) -> Void in
      return
    }, completion: { (context) -> Void in
      self.recalculateHipsterElementViews()
      for element in self.hipsterElements {
        element.view.hidden = false
      }
      self.addPaletteElements()
    })
  }

  // MARK: Actions

  @IBAction func selectPhotoTapped(sender: AnyObject) {
    let picker = UIImagePickerController()
    picker.delegate = self
    presentViewController(picker, animated: true, completion: nil)
  }

  func startEditingImage() {
    if let image = image {
      imageView.image = image

      autoHipsterizeCurrentImage()
      addPaletteElements()
    }
  }

  func autoHipsterizeCurrentImage() {
    // clear out the existing elements first
    for element in hipsterElements {
      element.view?.removeFromSuperview()
    }
    hipsterElements = []

    let hipsterize = Hipsterize(image: imageView.image!)
    for element in hipsterize.hipsterElements() {
      element.createViewForImageFrame(imageFrame)
      if let elementView = element.view {
        element.active = true
        view.addSubview(elementView)
      }

      hipsterElements.append(element)
    }

    addPaletteElements()
  }

  // MARK: - Gestures

  func handlePan(recognizer: UIPanGestureRecognizer) {
    switch recognizer.state {
    case .Began:
      if let element = closestElementViewForPoint(recognizer.locationInView(view), activeOnly: false) {
        if !element.active {
          // set things up for this element to join the party
          element.imageFrame = imageFrame
          // re-set the value to tickle the didSet observer once
          let scale = element.scale
          element.scale = scale
        }
        selectElement(element)
      }

    case .Changed:
      if (selectedElement == nil) { return }

      let translation = recognizer.translationInView(view)
      var newCenter = selectedElement!.center
      newCenter.x += translation.x
      newCenter.y += translation.y
      selectedElement!.center = newCenter

      recognizer.setTranslation(CGPoint.zeroPoint, inView: view)
    case .Cancelled, .Failed, .Ended:
      if (selectedElement == nil) { return }

      if elementIsInActiveRegion(selectedElement!) {
        selectedElement!.active = true
        editsWereMade = true
        addPaletteElements()
      } else {
        if selectedElement!.active {
          editsWereMade = true
        }

        selectedElement!.active = false

        // animate it away
        let index = (hipsterElements as NSArray).indexOfObject(selectedElement!)
        UIView.animateWithDuration(0.2, animations: {
          self.selectedElement!.view.alpha = 0
          }, completion: { (completed) -> Void in
            self.selectedElement!.view.removeFromSuperview()
            if index != NSNotFound {
              self.hipsterElements.removeAtIndex(index)
            } else {
              assertionFailure("Index of to-be-removed element not found")
            }
            self.selectedElement = nil
            self.addPaletteElements()
        })
      }
    default:
      break
    }
  }

  func handleTap(recognizer: UITapGestureRecognizer) {
    if let element = closestElementViewForPoint(recognizer.locationInView(view), activeOnly: true) {
      // tap on element
      selectElement(element)
    } else {
      if let element = selectedElement {
        // tap on background with something selected - deselecet
        deselectElement()
      } else {
        toggleUI()
      }
    }
  }

  func handlePinch(recognizer: UIPinchGestureRecognizer) {
    if let element = selectedElement {
      switch recognizer.state {
      case .Began:
        break
      case .Changed:
        element.scale = element.scale * recognizer.scale

        recognizer.scale = 1.0
      case .Cancelled, .Failed, .Ended:
        editsWereMade = true
      default:
        break
      }
    }
  }

  func handleRotation(recognizer: UIRotationGestureRecognizer) {
    if let element = selectedElement {
      switch recognizer.state {
      case .Began:
        break
      case .Changed:
        element.rotation += recognizer.rotation

        recognizer.rotation = 0
      case .Cancelled, .Failed, .Ended:
        editsWereMade = true
      default:
        break
      }
    }
  }

  // MARK: - UI

  func toggleUI() {
    // tap on background - toggle show/hide UI
    let setToHidden = self.paletteViewVisible()
    setUIHidden(setToHidden)
  }

  func setUIHidden(setToHidden: Bool) {
    self.navigationController?.setNavigationBarHidden(setToHidden, animated: true)
    UIView.animateWithDuration(0.2, animations: {
      self.paletteView.alpha = setToHidden ? 0 : 1
      for element in self.hipsterElements {
        if !element.active {
          element.view.alpha = setToHidden ? 0 : 1
        }
      }
    })
  }

  // MARK: - Element selection

  func selectElement(element: HipsterElement) {
    deselectElement()

    selectedElement = element
    selectedElement!.setGlow(true)
    view.bringSubviewToFront(selectedElement!.view)
  }

  func deselectElement() {
    selectedElement?.setGlow(false)
    selectedElement = nil
  }

  // MARK: - Calculation helpers

  private func recalculateHipsterElementViews() {
    for element in hipsterElements {
      element.setViewPositioningForImageFrame(imageFrame)
    }
  }

  func closestElementViewForPoint(point: CGPoint, activeOnly: Bool) -> HipsterElement? {
    var candidate: (distance: CGFloat, element: HipsterElement?) = (CGFloat.max, nil)

    func distanceBetweenPoints(p1: CGPoint, p2: CGPoint) -> CGFloat {
      return sqrt(pow(p2.x - p1.x, 2) + pow(p2.y - p1.y, 2))
    }

    for element in hipsterElements {
      if let elementView = element.view {
        if (!activeOnly || (activeOnly && element.active)) && CGRectContainsPoint(elementView.frame, point) {
          let centerDistance = distanceBetweenPoints(elementView.center, point)
          if centerDistance < candidate.distance {
            candidate.distance = centerDistance
            candidate.element = element
          }
        }
      }
    }

    return candidate.element
  }

  // MARK: - Palette elements

  func addPaletteElements() {
    var needs: [HipsterElementType: Bool] = [.Glasses1: true, .Glasses2: true, .Fedora1: true]

    for element in hipsterElements {
      if !element.active {
        needs[element.type] = false
      }
    }

    if needs[.Glasses1]! {
      let draggableElement = HipsterElement(type: .Glasses1, relativeCenter: CGPoint.zeroPoint, relativeWidth: CGRectGetWidth(glasses1.frame), rotation: 0)
      draggableElement.createViewWithFrame(view.convertRect(glasses1.bounds, fromView: glasses1))
      draggableElement.view.alpha = paletteViewVisible() ? 1 : 0
      view.addSubview(draggableElement.view)
      hipsterElements.append(draggableElement)
    }

    if needs[.Glasses2]! {
      let draggableElement = HipsterElement(type: .Glasses2, relativeCenter: CGPoint.zeroPoint, relativeWidth: CGRectGetWidth(glasses2.frame), rotation: 0)
      draggableElement.createViewWithFrame(view.convertRect(glasses2.bounds, fromView: glasses2))
      draggableElement.view.alpha = paletteViewVisible() ? 1 : 0
      view.addSubview(draggableElement.view)
      hipsterElements.append(draggableElement)
    }

    if needs[.Fedora1]! {
      let draggableElement = HipsterElement(type: .Fedora1, relativeCenter: CGPoint.zeroPoint, relativeWidth: CGRectGetWidth(fedora1.frame), rotation: 0)
      draggableElement.createViewWithFrame(view.convertRect(fedora1.bounds, fromView: fedora1))
      draggableElement.view.alpha = paletteViewVisible() ? 1 : 0
      view.addSubview(draggableElement.view)
      hipsterElements.append(draggableElement)
    }
  }

  func removePaletteElements() {
    var toRemove = [HipsterElement]()

    for element in hipsterElements {
      if !element.active {
        toRemove.append(element)
      }
    }

    for element in toRemove {
      let index = (hipsterElements as NSArray).indexOfObject(element)
      element.view.removeFromSuperview()

      if index != NSNotFound {
        hipsterElements.removeAtIndex(index)
      } else {
        assertionFailure("Index of to-be-removed element not found")
      }
    }
  }

  private func paletteViewVisible() -> Bool {
    return paletteView.alpha > 0
  }

  private func elementIsInActiveRegion(element: HipsterElement) -> Bool {
    if !paletteViewVisible() {
      return true
    } else {
      return !CGRectContainsPoint(paletteView.frame, element.view.center)
    }
  }
}

extension HipsterizeViewController: UIGestureRecognizerDelegate {
  func gestureRecognizer(gestureRecognizer: UIGestureRecognizer!, shouldRecognizeSimultaneouslyWithGestureRecognizer otherGestureRecognizer: UIGestureRecognizer!) -> Bool {
    return (gestureRecognizer == pinchRecognizer && otherGestureRecognizer == rotationRecognizer) || (gestureRecognizer == rotationRecognizer && otherGestureRecognizer == pinchRecognizer)
  }

  func gestureRecognizerShouldBegin(gestureRecognizer: UIGestureRecognizer) -> Bool {
    return true
  }
  
}

extension HipsterizeViewController: UIImagePickerControllerDelegate, UINavigationControllerDelegate {
  func imagePickerController(picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [NSObject : AnyObject]) {
    self.image = info[UIImagePickerControllerOriginalImage] as? UIImage
    startEditingImage()
    dismissViewControllerAnimated(true, completion: nil)
  }

  func imagePickerControllerDidCancel(picker: UIImagePickerController) {
    dismissViewControllerAnimated(true, completion: nil)
  }
}
